#!/usr/bin/env python
# -*- coding: utf8 -*-
# Plural-Forms for tr (Turkish)

# Turkish language has ONE form!
nplurals = 1

# Always returns 0:
get_plural_id = lambda n: 0

# Construct and return plural form of *word* using
# *plural_id* (which ALWAYS>0). This function will be executed
# for words (or phrases) not found in plural_dict dictionary
# construct_plural_form = lambda word, plural_id: word
